/*
 * Main.java
 *
 */
package splscli;

import com.seapine.licenseserver.api.LSAPI;
import com.seapine.licenseserver.api.GlobalUser;
import com.seapine.licenseserver.api.SerialNumber;
import com.seapine.licenseserver.api.Pool;
import com.seapine.licenseserver.api.SerialNumCounter;
import com.seapine.licenseserver.api.LSAPI.SecurityLevel;
import com.seapine.licenseserver.api.LSAPI.LicenseType;
import com.seapine.licenseserver.api.LSAPI.Product;
import java.lang.IllegalArgumentException;
import java.util.ArrayList;
import java.util.List;
import java.lang.StringBuffer;
import java.util.Arrays;
import java.util.Iterator;
import com.seapine.licenseserver.api.LicenseAssignment;

/**
 * Command line interface for LSAPI.  Demonstrates the Java wrapper with a java
 * cli implementation.
 */
public class Main {

    /**
     * Store username value specified on command line
     */
    private String username = "";
    /**
     * Store name value specified on command line
     */
    private String name = "";
    /**
     * Store ldap server value specified on command line
     */
    private String ldapServer = "";
    /**
     * Store TTPro license type value from command line
     */
    private String ttpro = "";
    /**
     * Store TTPro pool value from command line
     */
    private String ttpropool = "";
    /**
     * Store TCM license type value from command line
     */
    private String tcm = "";
    /**
     * Store TCM pool value from command line
     */
    private String tcmpool = "";
    /**
     * Store RM license type value from command line
     */
    private String ttrm = "";
    /**
     * Store RM pool value from command line
     */
    private String ttrmpool = "";
    /**
     * Store SCM license type value from command line
     */
    private String scm = "";
    /**
     * Store SCM pool value from command line
     */
    private String scmpool = "";
    /**
     * Store security level value from command line
     */
    private SecurityLevel securityLevel = SecurityLevel.noAccess;
    /**
     * username of user object in LDAP server
     */
    private String ldapUsername = "";
    /**
     * list of usernames specified at command line
     */
    private List<String> users;
    private String serialNumber = "";
    private long id = -1;
    private Product product = Product.ttp;
    private long fromID = -1;
    private long toID = -1;
    private long snID = -1;
    private int count = 0;
    /**
     * Store credentials and host info value from command line
     */
    private String credentials = null;
    /**
     * Store command value from command line
     */
    private String command = null;
    /**
     * Instance of LSAPI
     */
    private LSAPI api;

    /**
     * Arguments remaining after first pass
     */
    private List<String> remainingArgs;
    private static final String COMMAND_HELP = "HELP";
    private static final String COMMAND_VERSION = "VERSION";
    private static final String COMMAND_IMPORTUSER = "importuser";
    private static final String COMMAND_ADDUSER = "adduser";
    private static final String COMMAND_EDITUSER = "edituser";
    private static final String COMMAND_LISTUSERS = "listusers";
    private static final String COMMAND_ASSOCIATEUSER = "associateuser";
    private static final String COMMAND_ADDSN = "addsn";
    private static final String COMMAND_LISTSNS = "listsns";
    private static final String COMMAND_DELETESN = "deletesn";
    private static final String COMMAND_ADDPOOL = "addpool";
    private static final String COMMAND_EDITPOOL = "editpool";
    private static final String COMMAND_LISTPOOLS = "listpools";
    private static final String COMMAND_MOVELICENSES = "movelicenses";
    private static final String COMMAND_DELETEPOOL = "deletepool";

    /**
     * Creates a new instance of Main
     * @param args Command line arguments from execution
     */
    public Main(String[] args) {
        api = new LSAPI();
        remainingArgs = parseArgs(args);
    }

    /**
     * Parse command line arguments into internal state variables
     * @return list of arguments not parsed out
     * @param args List of command line arguments
     * @throws IllegalArgumentException for unrecognized parameters
     */
    private List<String> parseArgs(String[] args) throws IllegalArgumentException {
        List<String> remaining = new ArrayList<String>(args.length);

        for (int x = 0; x < args.length; x++) {
            if (args[x].charAt(0) != '-') {
                // only expected parameters without a hyphen are credentials and command...
                // credentials come first
                if (credentials == null) {
                    credentials = args[x];
                } else if (command == null) {
                    command = args[x];
                } else {
                    // malformed parameters
                    throw new IllegalArgumentException("Unexpected parameter: " + args[x]);

                }
            } else {
                // command modifiers
                if (args[x].equals("-h") || args[x].equals("--help")) {
                    command = COMMAND_HELP;
                } else if (args[x].equals("-v") || args[x].equals("--version")) {
                    command = COMMAND_VERSION;
                } // command data
                else if (args[x].contains("--username")) {
                    username = OptParser.splitOpt(args[x]);

                    // useful for parsing user
                    remaining.add(args[x]);
                } else if (args[x].contains("--name")) {
                    name = OptParser.splitOpt(args[x]);

                    // could be useful for parsing pool
                    remaining.add(args[x]);
                } else if (args[x].contains("--ldapserver")) {
                    ldapServer = OptParser.splitOpt(args[x]);
                } else if (args[x].contains("--ttpro")) {
                    ttpro = OptParser.splitOpt(args[x]);

                    // useful for parsing user
                    remaining.add(args[x]);
                } else if (args[x].contains("--ttpropool")) {
                    ttpropool = OptParser.splitOpt(args[x]);

                    // useful for parsing user
                    remaining.add(args[x]);
                } else if (args[x].contains("--tcm")) {
                    tcm = OptParser.splitOpt(args[x]);

                    // useful for parsing user
                    remaining.add(args[x]);
                } else if (args[x].contains("--tcmpool")) {
                    tcmpool = OptParser.splitOpt(args[x]);

                    // useful for parsing user
                    remaining.add(args[x]);
                } else if (args[x].contains("--ttrm")) {
                    ttrm = OptParser.splitOpt(args[x]);

                    // useful for parsing user
                    remaining.add(args[x]);
                } else if (args[x].contains("--ttrmpool")) {
                    ttrmpool = OptParser.splitOpt(args[x]);

                    // useful for parsing user
                    remaining.add(args[x]);
                } else if (args[x].contains("--scm")) {
                    scm = OptParser.splitOpt(args[x]);

                    // useful for parsing user
                    remaining.add(args[x]);
                } else if (args[x].contains("--scmpool")) {
                    scmpool = OptParser.splitOpt(args[x]);

                    // useful for parsing user
                    remaining.add(args[x]);
                } else if (args[x].contains("--lsusername")) {
                    username = OptParser.splitOpt(args[x]);
                } else if (args[x].contains("--ldapusername")) {
                    ldapUsername = OptParser.splitOpt(args[x]);
                } else if (args[x].contains("--security")) {
                    String secLevelOption = "";
                    secLevelOption = OptParser.splitOpt(args[x]);

                    if (secLevelOption.contains("none")) {
                        securityLevel = SecurityLevel.noAccess;
                    } else if (secLevelOption.contains("fetch")) {
                        securityLevel = SecurityLevel.retrieveUsers;
                    } else if (secLevelOption.contains("users")) {
                        securityLevel = SecurityLevel.manageUsers;
                    } else if (secLevelOption.contains("full")) {
                        securityLevel = SecurityLevel.fullControl;
                    } else {
                        securityLevel = SecurityLevel.noAccess;
                    }

                    // useful for parsing user
                    remaining.add(args[x]);
                } else if (args[x].contains("--users")) {
                    users = Arrays.asList(OptParser.splitOpt(args[x]).split(","));
                } else if (args[x].contains("--serialnumber")) {
                    serialNumber = OptParser.splitOpt(args[x]);
                } else if (args[x].contains("id")) {
                    id = Long.parseLong(OptParser.splitOpt(args[x]));
                    // could be useful for parsing pool
                    remaining.add(args[x]);
                } else if (args[x].contains("product")) {
                    String prodOpt = "";
                    prodOpt = OptParser.splitOpt(args[x]);

                    if (prodOpt.contains("ttpro")) {
                        product = Product.ttp;
                    } else if (prodOpt.contains("tcm")) {
                        product = Product.tcm;
                    } else if (prodOpt.contains("ttrm")) {
                        product = Product.rm;
                    } else if (prodOpt.contains("scm")) {
                        product = Product.scm;
                    }
                    // could be useful for parsing pool
                    remaining.add(args[x]);
                } else if (args[x].contains("fromid")) {
                    fromID = Long.parseLong(OptParser.splitOpt(args[x]));
                } else if (args[x].contains("toid")) {
                    toID = Long.parseLong(OptParser.splitOpt(args[x]));
                } else if (args[x].contains("snid")) {
                    snID = Long.parseLong(OptParser.splitOpt(args[x]));
                } else if (args[x].contains("count")) {
                    count = Integer.parseInt(OptParser.splitOpt(args[x]));
                } else {
                    // could be useful for parsing user or pool
                    remaining.add(args[x]);
                }
            }
        }

        return remaining;
    }

    /**
     * Execute the command specified in command line parameters
     * @throws java.lang.Exception A number of exceptions can be thrown by this function
     */
    public void executeCommand() throws Exception {
        List<LicenseAssignment> licenses = null;

        if (command == null) {
            throw new IllegalArgumentException("No command specified");
        }
        if (command.equals(COMMAND_IMPORTUSER)) {
            LSCredentials creds = new LSCredentials(credentials);
            List<LicenseType> licenseTypes = OptParser.parseLicenses(ttpro, tcm, ttrm, scm);

            api.connect(creds.hostname(), creds.port(), creds.username(), creds.password());
            api.importLDAPUser(name, username, ldapServer, securityLevel, licenseTypes);
            api.disconnect();
        } else if (command.equals(COMMAND_ADDUSER)) {
            LSCredentials creds = new LSCredentials(credentials);

            api.connect(creds.hostname(), creds.port(), creds.username(), creds.password());

            GlobalUser user = new ParsedGlobalUser(remainingArgs);
            api.addUserEx(user);
            api.disconnect();
        } else if (command.equals(COMMAND_EDITUSER)) {
            LSCredentials creds = new LSCredentials(credentials);
            licenses = OptParser.parseLicenses(ttpro, ttpropool, tcm, tcmpool, ttrm, ttrmpool, scm,
                    scmpool);

            api.connect(creds.hostname(), creds.port(), creds.username(), creds.password());

            // get the user, apply any changes...
            List<String> usernames = new ArrayList<String>();
            usernames.add(username);
            List<GlobalUser> globalUsers = api.getUsers(usernames);

            if (globalUsers.size() < 1) {
                throw new IllegalArgumentException("Could not find user " + username);
            } else if (globalUsers.size() > 1) {
                throw new IllegalArgumentException("Did not specify a unique user");
            }

            GlobalUser user = new ParsedGlobalUser(globalUsers.get(0), remainingArgs);

            api.editUser(username, user);
            api.disconnect();
        } else if (command.equals(COMMAND_LISTUSERS)) {
            LSCredentials creds = new LSCredentials(credentials);

            api.connect(creds.hostname(), creds.port(), creds.username(), creds.password());
            List<GlobalUser> globalUsers = api.getUsers(users);

            System.out.println(GlobalUser.displayHeader());
            Iterator<GlobalUser> iter = globalUsers.iterator();
            while (iter.hasNext()) {
                System.out.println(iter.next().toString());
            }
            api.disconnect();

        } else if (command.equals(COMMAND_ASSOCIATEUSER)) {
            LSCredentials creds = new LSCredentials(credentials);
            licenses = OptParser.parseLicenses(ttpro, ttpropool, tcm, tcmpool, ttrm, ttrmpool, scm,
                    scmpool);

            api.connect(creds.hostname(), creds.port(), creds.username(), creds.password());
            api.associateToLDAPUser(username, ldapUsername, ldapServer);
            api.disconnect();

        } else if (command.equals(COMMAND_ADDSN)) {
            LSCredentials creds = new LSCredentials(credentials);

            api.connect(creds.hostname(), creds.port(), creds.username(), creds.password());
            api.addSerialNumber(serialNumber);
            api.disconnect();
        } else if (command.equals(COMMAND_LISTSNS)) {
            LSCredentials creds = new LSCredentials(credentials);

            api.connect(creds.hostname(), creds.port(), creds.username(), creds.password());
            List<SerialNumber> sns = api.getSerialNumbers();

            System.out.println(SerialNumber.displayHeader());
            Iterator<SerialNumber> iter = sns.iterator();
            while (iter.hasNext()) {
                System.out.println(iter.next().toString());
            }
            api.disconnect();
        } else if (command.equals(COMMAND_DELETESN)) {
            LSCredentials creds = new LSCredentials(credentials);

            api.connect(creds.hostname(), creds.port(), creds.username(), creds.password());
            api.deleteSerialNumber(id);
            api.disconnect();
        } else if (command.equals(COMMAND_ADDPOOL)) {
            LSCredentials creds = new LSCredentials(credentials);
            Pool newPool = new ParsedPool(remainingArgs);

            api.connect(creds.hostname(), creds.port(), creds.username(), creds.password());
            api.addPool(newPool);
            api.disconnect();
        } else if (command.equals(COMMAND_LISTPOOLS)) {
            LSCredentials creds = new LSCredentials(credentials);

            api.connect(creds.hostname(), creds.port(), creds.username(), creds.password());
            List<Pool> pools = api.getPools();

            System.out.println(Pool.displayHeader());
            Iterator<Pool> iter = pools.iterator();
            while (iter.hasNext()) {
                System.out.println(iter.next().toString());
            }
            api.disconnect();
        } else if (command.equals(COMMAND_EDITPOOL)) {
            LSCredentials creds = new LSCredentials(credentials);

            api.connect(creds.hostname(), creds.port(), creds.username(), creds.password());
            List<Pool> pools = api.getPools();
            Pool poolToEdit = null;
            Iterator<Pool> iter = pools.iterator();
            while (iter.hasNext()) {
                if (iter.next().id == id) {
                    poolToEdit = iter.next();
                }
            }

            if (poolToEdit != null) {
                List<Pool> poolsToSave = new ArrayList<Pool>(1);
                poolsToSave.add(new ParsedPool(poolToEdit, remainingArgs));
                api.editPools(poolsToSave);
            } else {
                throw new IllegalArgumentException("Could not find pool with id " + id);
            }

            api.disconnect();
        } else if (command.equals(COMMAND_MOVELICENSES)) {
            LSCredentials creds = new LSCredentials(credentials);

            api.connect(creds.hostname(), creds.port(), creds.username(), creds.password());

            // find the pools to be edited
            List<Pool> pools = api.getPools();
            List<Pool> editPools = new ArrayList<Pool>(2);
            final int FROM = 0;
            final int TO = 1;
            Iterator<Pool> iter = pools.iterator();
            while (iter.hasNext()) {
                if (iter.next().id == fromID) {
                    editPools.add(FROM, new Pool(iter.next()));
                } else if (iter.next().id == toID) {
                    editPools.add(TO, new Pool(iter.next()));
                }
            }

            if (editPools.size() != 2) {
                throw new IllegalArgumentException("A specified pool does not exist in License Server");
            } else {
                boolean notFound = true;
                boolean tooMany = false;
                SerialNumCounter remove = null;
                Iterator<SerialNumCounter> fromAsgn = editPools.get(FROM).serialNumbers.iterator();
                while (fromAsgn.hasNext()) {
                    if (fromAsgn.next().serialNumberID == snID) {
                        notFound = false;

                        if (fromAsgn.next().count > count) {
                            fromAsgn.next().count -= count;
                        } else if (fromAsgn.next().count == count) {
                            remove = fromAsgn.next();
                        } else {
                            // trying to remove more than are allocated
                            tooMany = true;
                        }
                    }
                }

                if (tooMany) {
                    throw new IllegalArgumentException("Attempted to remove more licenses that configured from serial number with id" + snID);
                } else if (notFound) {
                    throw new IllegalArgumentException("The source pool does not contain licenses from serial number with id " + snID);
                }

                if (remove != null) {
                    editPools.get(FROM).serialNumbers.remove(remove);
                }

                notFound = true;
                Iterator<SerialNumCounter> toAsgn = editPools.get(TO).serialNumbers.iterator();
                while (toAsgn.hasNext()) {
                    if (toAsgn.next().serialNumberID == snID) {
                        notFound = false;
                        toAsgn.next().count += count;
                    }
                }

                if (notFound) {
                    editPools.get(TO).serialNumbers.add(new SerialNumCounter(snID, count));
                }

                api.editPools(editPools);
            }
            api.disconnect();
        } else if (command.equals(COMMAND_DELETEPOOL)) {
            LSCredentials creds = new LSCredentials(credentials);

            api.connect(creds.hostname(), creds.port(), creds.username(), creds.password());
            api.deletePool(id);
            api.disconnect();
        } else if (command.equals(COMMAND_HELP)) {
            printHelp();
        } else if (command.equals(COMMAND_VERSION)) {
            System.out.println("splscli: Helix ALM License Server Command Line Tool 1.1.0");
        } else {
            throw new IllegalArgumentException("Unrecognized command: " + command);
        }
    }

    /**
     * Main routine for splscli
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        try {
            Main app = new Main(args);

            app.executeCommand();
        } catch (java.lang.Exception ex) {
            System.out.println(ex.getMessage());
            Main.printHelp();

            ex.printStackTrace();
        } catch (java.lang.Error er) {
            System.out.println(er.getMessage());
            Main.printHelp();

            er.printStackTrace();
        }
    }

    /**
     * Write usage message to console
     */
    public static void printHelp() {
        System.out.println("Usage: splscli [[username[:password]]@hostname[:port]] [command] [options]");
        System.out.println("Commands:");
        System.out.println("  importuser  Import a user from LDAP");
        System.out.println("     --username    pattern to match user's username");
        System.out.println("     --name        pattern to match user's name");
        System.out.println("     --ldapserver  ldap server from which to import user");
        System.out.println("     --security    License Server security level (none, fetch, users, full)");
        System.out.println("     --ttpro       TestTrack Pro license type (none, float, named)");
        System.out.println("     --tcm         TestTrack TCM license type (none, float, named)");
        System.out.println("     --ttrm        TestTrack RM license type (none, float, named)");
        System.out.println("     --scm         Surround SCM license type (none, float, named)");
        System.out.println("  associateuser  Associate a License Server user to an LDAP user\n");
        System.out.println("     --lsusername             user's username in License Server\n");
        System.out.println("     --ldapusername           user's username in LDAP (if not specified, lsusername will be used)\n");
        System.out.println("     --ldapserver             LDAP server in which to associate user\n");
        System.out.println("  adduser     Add a user to License Server");
        System.out.println("     --firstname   user's first name");
        System.out.println("     --initials    user's initials");
        System.out.println("     --lastname    user's last name");
        System.out.println("     --username    user's username");
        System.out.println("     --password    user's password");
        System.out.println("     --active                 whether user is active or inactive (yes, no)\n");
        System.out.println("     --emailtype              type of user's email address (smtp, mapi, other)\n");
        System.out.println("     --email       user's email address");
        System.out.println("     --phone1type             type of user's phone number (work, fax, home, pager, mobile)\n");
        System.out.println("     --phone1                 user's phone number\n");
        System.out.println("     --phone2type             type of user's phone number (work, fax, home, pager, mobile)\n");
        System.out.println("     --phone2                 user's phone number\n");
        System.out.println("     --forcechangepassword    whether the user must change his password at next login (yes, no)\n");
        System.out.println("     --preventchangepassword  whether the user is prevented from changing his password (yes, no)\n");
        System.out.println("     --type                   type of user (user, customer)\n");
        System.out.println("     --address     user's address");
        System.out.println("     --company     user's company");
        System.out.println("     --division    user's division");
        System.out.println("     --department  user's department");
        System.out.println("     --security    License Server security level (none, fetch, users, full)");
        System.out.println("     --singlesignon           whether the user is permitted to login using single sign-on (yes, no)\n");
        System.out.println("     --ttpro       TestTrack Pro license type (none, float, named)");
        System.out.println("     --ttpropool   TestTrack Pro floating license pool");
        System.out.println("     --tcm         TestTrack TCM license type (none, float, named)");
        System.out.println("     --tcmpool                TestTrack TCM floating license pool");
        System.out.println("     --ttrm        TestTrack RM license type (none, float, named)");
        System.out.println("     --ttrmpool               TestTrack RM floating license pool");
        System.out.println("     --scm         Surround SCM license type (none, float, named)");
        System.out.println("     --scmpool                Surround SCM floating license pool");
        System.out.println("     --notes                  user notes\n");
        System.out.println("  listusers      List all users in License Server\n");
        System.out.println("     --users                  comma-delimited list of usernames to list\n");
        System.out.println("  edituser       Modify a user in License Server\n");
        System.out.println("     --username               username of user to modify\n");
        System.out.println("     --firstname   user's first name");
        System.out.println("     --initials    user's initials");
        System.out.println("     --lastname    user's last name");
        System.out.println("     --newusername            new username for user\n");
        System.out.println("     --password    user's password");
        System.out.println("     --active                 whether user is active or inactive (yes, no)\n");
        System.out.println("     --emailtype              type of user's email address (smtp, mapi, other)\n");
        System.out.println("     --email       user's email address");
        System.out.println("     --phone1type             type of user's phone number (work, fax, home, pager, mobile)\n");
        System.out.println("     --phone1                 user's phone number\n");
        System.out.println("     --phone2type             type of user's phone number (work, fax, home, pager, mobile)\n");
        System.out.println("     --phone2                 user's phone number\n");
        System.out.println("     --forcechangepassword    whether the user must change his password at next login (yes, no)\n");
        System.out.println("     --preventchangepassword  whether the user is prevented from changing his password (yes, no)\n");
        System.out.println("     --type                   type of user (user, customer)\n");
        System.out.println("     --address     user's address");
        System.out.println("     --company     user's company");
        System.out.println("     --division    user's division");
        System.out.println("     --department  user's department");
        System.out.println("     --security    License Server security level (none, fetch, users, full)");
        System.out.println("     --singlesignon           whether the user is permitted to login using single sign-on (yes, no)\n");
        System.out.println("     --ttpro       TestTrack Pro license type (none, float, named)");
        System.out.println("     --ttpropool   TestTrack Pro floating license pool");
        System.out.println("     --tcm         TestTrack TCM license type (none, float, named)");
        System.out.println("     --tcmpool                TestTrack TCM floating license pool");
        System.out.println("     --ttrm        TestTrack RM license type (none, float, named)");
        System.out.println("     --ttrmpool               TestTrack RM floating license pool");
        System.out.println("     --scm         Surround SCM license type (none, float, named)");
        System.out.println("     --scmpool                Surround SCM floating license pool");
        System.out.println("     --notes                  user notes");
        System.out.println("  addsn          Add a serial number to License Server.");
        System.out.println("     --serialnumber           serial number to add to License Server");
        System.out.println("  listsns        Display a list of serial number information.  Each line is tab-delimited.");
        System.out.println("  deletesn       Remove a serial number from License Server.");
        System.out.println("     --id                     unique identifier of serial number to remove from License Server");
        System.out.println("  addpool        Add a pool to License Server.");
        System.out.println("    --name                    pool's name");
        System.out.println("    --product                 product for which pool will contain licenses (ttpro, tcm, ttrm, scm)");
        System.out.println("  listpools      Display list of pool information.  Each line is tab-delimited.");
        System.out.println("  editpool       Modify pool in License Server.");
        System.out.println("    --id                      unique identifer of pool to modify");
        System.out.println("    --name                    new name for pool");
        System.out.println("    --product                 new product for pool (ttpro, tcm, ttrm, scm)  Note: changing product will move all licenses to the default pool.");
        System.out.println("  movelicenses   Move licenses between pools in License Server");
        System.out.println("    --fromid                  unique identifier of pool from which to remove licenses");
        System.out.println("    --toid                    unique identifier of pool to which to add licenses");
        System.out.println("    --snid                    unique identifier of serial number for which to move licenses");
        System.out.println("    --count                   number of licenses to move");
        System.out.println("  deletepool     Remove a pool from License Server.");
        System.out.println("    --id                      unique identifier of pool to delete");
        System.out.println("  -h  --help");
        System.out.println("  -V  --version");
    }
}
