#include "lsapi.h"
#include <getopt.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include <assert.h>

#define MIN(a,b) ((a) < (b) ? (a) : (b))

static const char *programName;

static void printUsage(FILE *outStream)
{
   fprintf(outStream, "Usage: %s [[username[:password]]@hostname[:port]] [command] [options]\n",
           programName);
   fprintf(outStream, "Commands:\n");
   fprintf(outStream, "  importuser     Import a user from LDAP\n"
                      "     --username               pattern to match user's username\n"
                      "     --name                   pattern to match user's name\n"
                      "     --ldapserver             LDAP server from which to import user\n"
                      "     --security               License Server security level (none, fetch, users, full)\n"
                      "     --ttpro                  TestTrack Pro license type (none, float, named)\n"
                      "     --tcm                    TestTrack TCM license type (none, float, named)\n");
   fprintf(outStream, "     --ttrm                   TestTrack RM license type (none, float, named)\n"
                      "     --scm                    Surround SCM license type (none, float, named)\n"
                      "  associateuser  Associate a License Server user to an LDAP user\n"
                      "     --lsusername             user's username in License Server\n"
                      "     --ldapusername           user's username in LDAP (if not specified, lsusername will be used)\n"
                      "     --ldapserver             LDAP server in which to associate user\n"
                      "  adduser        Add a user to License Server\n");
   fprintf(outStream, "     --firstname              user's first name\n"
                      "     --initials               user's initials\n"
                      "     --lastname               user's last name\n"
                      "     --username               user's username\n"
                      "     --password               user's password\n"
                      "     --active                 whether user is active or inactive (yes, no)\n"
                      "     --emailtype              type of user's email address (smtp, mapi, other)\n"
                      "     --email                  user's email address\n");
   fprintf(outStream, "     --phone1type             type of user's phone number (work, fax, home, pager, mobile)\n"
                      "     --phone1                 user's phone number\n"
                      "     --phone2type             type of user's phone number (work, fax, home, pager, mobile)\n"
                      "     --phone2                 user's phone number\n"
                      "     --forcechangepassword    whether the user must change his password at next login (yes, no)\n"
                      "     --preventchangepassword  whether the user is prevented from changing his password (yes, no)\n");
   fprintf(outStream, "     --type                   type of user (user, customer)\n"
                      "     --address                user's address\n"
                      "     --company                user's company\n"
                      "     --division               user's division\n"
                      "     --department             user's department\n"
                      "     --security               License Server security level (none, fetch, users, full)\n"
                      "     --singlesignon           whether the user is permitted to login using single sign-on (yes, no)\n"
                      "     --ttpro                  TestTrack Pro license type (none, float, named)\n");
   fprintf(outStream, "     --ttpropool              TestTrack Pro floating license pool\n"
                      "     --tcm                    TestTrack TCM license type (none, float, named)\n"
                      "     --tcmpool                TestTrack TCM floating license pool\n"
                      "     --ttrm                   TestTrack RM license type (none, float, named)\n"
                      "     --ttrmpool               TestTrack RM floating license pool\n"
                      "     --scm                    Surround SCM license type (none, float, named)\n"
                      "     --scmpool                Surround SCM floating license pool\n");
   fprintf(outStream, "     --notes                  user notes\n"
                      "  listusers      Display list of user information.  Each line is tab-delimited, the notes field is quoted.\n"
                      "     --users                  comma-delimited list of usernames to list\n"
                      "  edituser       Modify a user in License Server\n");
   fprintf(outStream, "     --username               username of user to modify\n"
                      "     --firstname              user's first name\n"
                      "     --initials               user's initials\n"
                      "     --lastname               user's last name\n"
                      "     --newusername            new username for user\n"
                      "     --password               user's password\n"
                      "     --active                 whether user is active or inactive (yes, no)\n");
   fprintf(outStream, "     --emailtype              type of user's email address (smtp, mapi, other)\n"
                      "     --email                  user's email address\n"
                      "     --phone1type             type of user's phone number (work, fax, home, pager, mobile)\n"
                      "     --phone1                 user's phone number\n"
                      "     --phone2type             type of user's phone number (work, fax, home, pager, mobile)\n"
                      "     --phone2                 user's phone number\n"
                      "     --forcechangepassword    whether the user must change his password at next login (yes, no)\n");
   fprintf(outStream, "     --preventchangepassword  whether the user is prevented from changing his password (yes, no)\n"
                      "     --type                   type of user (user, customer)\n"
                      "     --address                user's address\n"
                      "     --company                user's company\n"
                      "     --division               user's division\n"
                      "     --department             user's department\n"
                      "     --security               License Server security level (none, fetch, users, full)\n");
   fprintf(outStream, "     --singlesignon           whether the user is permitted to login using single sign-on (yes, no)\n"
                      "     --ttpro                  TestTrack Pro license type (none, float, named)\n"
                      "     --ttpropool              TestTrack Pro floating license pool\n"
                      "     --tcm                    TestTrack TCM license type (none, float, named)\n"
                      "     --tcmpool                TestTrack TCM floating license pool\n"
                      "     --ttrm                   TestTrack RM license type (none, float, named)\n");
   fprintf(outStream, "     --ttrmpool               TestTrack RM floating license pool\n"
                      "     --scm                    Surround SCM license type (none, float, named)\n"
                      "     --scmpool                Surround SCM floating license pool\n"
                      "     --notes                  user notes\n");
   fprintf(outStream, "  addsn          Add a serial number to License Server.\n"
                      "     --serialnumber           serial number to add to License Server\n");
   fprintf(outStream, "  listsns        Display a list of serial number information.  Each line is tab-delimited.\n");
   fprintf(outStream, "  deletesn       Remove a serial number from License Server.\n"
                      "     --id                     unique identifier of serial number to remove from License Server\n");
   fprintf(outStream, "  addpool        Add a pool to License Server.\n"
                      "    --name                    pool's name\n"
                      "    --product                 product for which pool will contain licenses (ttpro, tcm, ttrm, scm)\n");
   fprintf(outStream, "  listpools      Display list of pool information.  Each line is tab-delimited.\n");
   fprintf(outStream, "  editpool       Modify pool in License Server.\n"
                      "    --id                      unique identifer of pool to modify\n"
                      "    --name                    new name for pool\n"
                      "    --product                 new product for pool (ttpro, tcm, ttrm, scm)  Note: changing product will move all licenses to the default pool.\n");
   fprintf(outStream, "  movelicenses   Move licenses between pools in License Server\n"
                      "    --fromid                  unique identifier of pool from which to remove licenses\n"
                      "    --toid                    unique identifier of pool to which to add licenses\n"
                      "    --snid                    unique identifier of serial number for which to move licenses\n"
                      "    --count                   number of licenses to move\n");
   fprintf(outStream, "  deletepool     Remove a pool from License Server.\n"
                      "    --id                      unique identifier of pool to delete\n");
   fprintf(outStream, "  -h  --help\n"
                      "  -V  --version\n");
}

static int parseLoginInfo(const char *lsLoginInfo, char **username, char **password,
                          char **hostname, int *port)
{
   /* login information comes in the following format:
    *    username:password@hostname:port
    * the following portions are optional:
    *   - username:password
    *   - port
    * platforms that support single sign-on will not require username/password
    * if no port is specified, the default of 5100 will be used
    * hostname is always required
    */
   int retVal = 1;
   char *localLoginInfo = NULL;
   char *hostinfo = NULL;
   char *credentials = NULL;
   char *localUsername = NULL;
   char *localPassword = NULL;
   char *localHostName = NULL;
   char *localPort = NULL;

   if(lsLoginInfo)
   {
      localLoginInfo = (char*)calloc(strlen(lsLoginInfo) + 1, sizeof(char));
      strcpy(localLoginInfo, lsLoginInfo);

      /* split the string at the '@' */
      if((credentials = strtok(localLoginInfo, "@")))
      {
         if(!(hostinfo = strtok(NULL, "@")))
         {
            hostinfo = credentials;
            credentials = NULL;
         }
      }

      /* divide credentials up by ':' */
      if(credentials)
      {
         if((localUsername = strtok(credentials, ":")))
         {
            localPassword = strtok(NULL, ":");
         }
         else
         {
            localUsername = credentials;
         }
      }

      /* divide host info up by ':' */
      if(hostinfo)
      {
         if((localHostName = strtok(hostinfo, ":")))
         {
            localPort = strtok(NULL, ":");
         }
         else
         {
            localHostName = hostinfo;
         }
      }

      /* copy out username */
      if(username)
      {
         if(localUsername)
         {
            *username = (char*)calloc(strlen(localUsername) + 1, sizeof(char));
            if(*username)
            {
               strcpy(*username, localUsername);
            }
         }
         else
         {
            *username = (char*)calloc(1, sizeof(char));
         }
      }

      /* copy out password */
      if(password)
      {
         if(localPassword)
         {
            *password = (char*)calloc(strlen(localPassword) + 1, sizeof(char));
            if(*password)
            {
               strcpy(*password, localPassword);
            }
         }
         else
         {
            *password = (char*)calloc(1, sizeof(char));
         }
      }

      /* copy out host name*/
      if(hostname)
      {
         if(localHostName)
         {
            *hostname = (char*)calloc(strlen(localHostName) + 1, sizeof(char));
            if(*hostname)
            {
               strcpy(*hostname, localHostName);
            }
         }
         else
         {
            *hostname = (char*)calloc(1, sizeof(char));
         }
      }

      /* copy out port */
      if(port)
      {
         if(localPort)
         {
            *port = atoi(localPort);
         }
         else
         {
            *port = 5100;
         }
      }

      free(localLoginInfo);
   }

   if(hostname && *hostname && strlen(*hostname) > 0)
   {
      /* the only thing that is absolutely required is a host name */
      retVal = 0;
   }

   return retVal;
}

static void parseSecurity(const char *securityLevel, int *security)
{
   int securityLevelLen = 0;
   if(securityLevel && security)
   {
      securityLevelLen = strlen(securityLevel);
      if(strncmp("none", securityLevel, MIN(4, securityLevelLen)) == 0)
      {
         (*security) = kNoAccess;
      }
      else if(strncmp("fetch", securityLevel, MIN(5, securityLevelLen)) == 0)
      {
         (*security) = kRetrieveUsers;
      }
      else if(strncmp("users", securityLevel, MIN(5, securityLevelLen)) == 0)
      {
         (*security) = kManageUsers;
      }
      else if(strncmp("full", securityLevel, MIN(4, securityLevelLen)) == 0)
      {
         (*security) = kFullControl;
      }
      else
      {
         /* default: no access */
         (*security) = kNoAccess;
      }
   }
}

static const char* getSecurity(int security)
{
   switch(security)
   {
      case kNoAccess:
         return "none";
      case kRetrieveUsers:
         return "fetch";
      case kManageUsers:
         return "users";
      case kFullControl:
         return "full";
   }

   return "none";
}

#define PRODUCT_TTP 0
#define PRODUCT_TCM 1
#define PRODUCT_RM 2
#define PRODUCT_SCM 3
#define NUM_PRODUCTS 4

static void parseLicensing(const char *ttpro, const char *ttpropool, const char *tcm,
                           const char *tcmpool, const char *ttrm, const char *ttrmpool,
                           const char *scm, const char *scmpool, LicenseAssignment ***licenses,
                           int *numLicenses)
{
   int curLicense = 0;
   int x = 0;
   int licenseData[NUM_PRODUCTS];
   unsigned int poolData[NUM_PRODUCTS];
   memset(licenseData, 0, sizeof(int)*NUM_PRODUCTS);
   memset(poolData, 0, sizeof(unsigned int)*NUM_PRODUCTS);

   if(licenses && numLicenses)
   {
      /* if *licenses already exists, preserve data from it */
      if(*licenses && *numLicenses > 0)
      {
         for(x = 0; x < *numLicenses; x++)
         {
            if((*licenses)[x]->licenseType == kTTPFloating ||
               (*licenses)[x]->licenseType == kTTPNamed)
            {
               licenseData[PRODUCT_TTP] = (*licenses)[x]->licenseType;
               poolData[PRODUCT_TTP] = (*licenses)[x]->associatedID;
            }
            else if((*licenses)[x]->licenseType == kTCMFloating ||
                    (*licenses)[x]->licenseType == kTCMNamed)
            {
               licenseData[PRODUCT_TCM] = (*licenses)[x]->licenseType;
               poolData[PRODUCT_TCM] = (*licenses)[x]->associatedID;
            }
            else if((*licenses)[x]->licenseType == kRMFloating ||
                    (*licenses)[x]->licenseType == kRMNamed)
            {
               licenseData[PRODUCT_RM] = (*licenses)[x]->licenseType;
               poolData[PRODUCT_RM] = (*licenses)[x]->associatedID;
            }
            else if((*licenses)[x]->licenseType == kSCMFloating ||
                    (*licenses)[x]->licenseType == kSCMNamed)
            {
               licenseData[PRODUCT_SCM] = (*licenses)[x]->licenseType;
               poolData[PRODUCT_SCM] = (*licenses)[x]->associatedID;
            }
         }

         free(*licenses);
      }

      if(ttpro)
      {
         if(strncmp("float", ttpro, MIN(5, strlen(ttpro))) == 0)
         {
            licenseData[PRODUCT_TTP] = kTTPFloating;
         }
         else if(strncmp("named", ttpro, MIN(5, strlen(ttpro))) == 0)
         {
            licenseData[PRODUCT_TTP] = kTTPNamed;
         }
         else if(strncmp("none", ttpro, MIN(4, strlen(ttpro))) == 0)
         {
            licenseData[PRODUCT_TTP] = 0;
         }
      }

      if(ttpropool)
      {
         if(licenseData[PRODUCT_TTP] == kTTPFloating)
         {
            sscanf(ttpropool, "%u", &poolData[PRODUCT_TTP]);
         }
      }

      if(tcm)
      {
         if(strncmp("float", tcm, MIN(5, strlen(tcm))) == 0)
         {
            licenseData[PRODUCT_TCM] = kTCMFloating;
         }
         else if(strncmp("named", tcm, MIN(5, strlen(tcm))) == 0)
         {
            licenseData[PRODUCT_TCM] = kTCMNamed;
         }
         else if(strncmp("none", tcm, MIN(4, strlen(tcm))) == 0)
         {
            licenseData[PRODUCT_TCM] = 0;
         }
      }

      if(tcmpool)
      {
         if(licenseData[PRODUCT_TCM] == kTCMFloating)
         {
            sscanf(tcmpool, "%u", &poolData[PRODUCT_TCM]);
         }
      }

      if(ttrm)
      {
         if(strncmp("float", ttrm, MIN(5, strlen(ttrm))) == 0)
         {
            licenseData[PRODUCT_RM] = kRMFloating;
         }
         else if(strncmp("named", ttrm, MIN(5, strlen(ttrm))) == 0)
         {
            licenseData[PRODUCT_RM] = kRMNamed;
         }
         else if(strncmp("none", ttrm, MIN(4, strlen(ttrm))) == 0)
         {
            licenseData[PRODUCT_RM] = 0;
         }
      }

      if(ttrmpool)
      {
         if(licenseData[PRODUCT_RM] == kRMFloating)
         {
            sscanf(ttrmpool, "%u", &poolData[PRODUCT_RM]);
         }
      }

      if(scm)
      {
         if(strncmp("float", scm, MIN(5, strlen(scm))) == 0)
         {
            licenseData[PRODUCT_SCM] = kSCMFloating;
         }
         else if(strncmp("named", scm, MIN(5, strlen(scm))) == 0)
         {
            licenseData[PRODUCT_SCM] = kSCMNamed;
         }
         else if(strncmp("none", scm, MIN(4, strlen(scm))) == 0)
         {
            licenseData[PRODUCT_SCM] = 0;
         }
      }

      if(scmpool)
      {
         if(licenseData[PRODUCT_SCM] == kSCMFloating)
         {
            sscanf(scmpool, "%u", &poolData[PRODUCT_SCM]);
         }
      }

      (*numLicenses) = 0;

      for(x = 0; x < NUM_PRODUCTS; x++)
      {
         if(licenseData[x] > 0)
         {
            (*numLicenses)++;
         }
      }

      if((*numLicenses) > 0)
      {
         *licenses = (LicenseAssignment**)calloc(*numLicenses, sizeof(LicenseAssignment**));

         for(x = 0; x < NUM_PRODUCTS; x++)
         {
            if(licenseData[x] > 0)
            {
               (*licenses)[curLicense++] = (LicenseAssignment*)calloc(1, sizeof(LicenseAssignment*));
               (*licenses)[curLicense++]->licenseType = licenseData[x];
               (*licenses)[curLicense++]->associatedID = poolData[x];
            }
         }
      }
      else
      {
         (*licenses) = NULL;
      }
   }
}

static const char* getLicense(int license)
{
   switch(license)
   {
      case kTTPFloating:
         return "ttpro(float)";
         break;
      case kTTPNamed:
         return "ttpro(named";
         break;
      case kSCMFloating:
         return "scm(float)";
         break;
      case kSCMNamed:
         return "scm(named)";
         break;
      case kTCMFloating:
         return "tcm(float)";
         break;
      case kTCMNamed:
         return "tcm(named)";
         break;
      case kRMFloating:
         return "ttrm(float)";
         break;
      case kRMNamed:
         return "ttrm(named)";
         break;
   }
   return "";
}

static char* getLicenses(LicenseAssignment *licenses[], int numLicenses)
{
   /* longest single license string: "ttpro(float)," */
   int maxLen = 13 * numLicenses;
   char *licenseText = calloc(maxLen + 1, sizeof(char));
   int x = 0;
   int remainingLen = maxLen;

   if(licenseText)
   {
      if(licenses && numLicenses > 0)
      {
         for(x = 0; x < numLicenses - 1; x++)
         {
            strncat(licenseText, getLicense(licenses[x]->licenseType), remainingLen);
            remainingLen = maxLen - strlen(licenseText);
            strncat(licenseText, ",", remainingLen);
            remainingLen--;
         }

         strncat(licenseText, getLicense(licenses[numLicenses - 1]->licenseType), remainingLen);
      }
   }

   return licenseText;
}

static void parseEmailType(const char *emailType, int *outEmailType)
{
   if(emailType && outEmailType)
   {
      if(strncmp("smtp", emailType, MIN(4, strlen(emailType))) == 0)
      {
         *outEmailType = kInternetEmail;
      }
      else if(strncmp("mapi", emailType, MIN(4, strlen(emailType))) == 0)
      {
         *outEmailType = kMAPIEmail;
      }
      else if(strncmp("other", emailType, MIN(5, strlen(emailType))) == 0)
      {
         *outEmailType = kOtherEmail;
      }
      else
      {
         /* default: smpt */
         *outEmailType = kInternetEmail;
      }
   }
}

static const char* getEmailType(int emailType)
{
   switch(emailType)
   {
      case kInternetEmail:
         return "smtp";
      case kMAPIEmail:
         return "mapi";
      case kOtherEmail:
         return "other";
   }
   return "smtp";
}

static void parsePhoneType(const char *phoneType, int *outPhoneType)
{
   if(phoneType && outPhoneType)
   {
      if(strncmp("work", phoneType, MIN(4, strlen(phoneType))) == 0)
      {
         *outPhoneType = kWorkPhone;
      }
      else if(strncmp("fax", phoneType, MIN(3, strlen(phoneType))) == 0)
      {
         *outPhoneType = kFaxPhone;
      }
      else if(strncmp("home", phoneType, MIN(4, strlen(phoneType))) == 0)
      {
         *outPhoneType = kHomePhone;
      }
      else if(strncmp("pager", phoneType, MIN(5, strlen(phoneType))) == 0)
      {
         *outPhoneType = kPagerPhone;
      }
      else if(strncmp("mobile", phoneType, MIN(6, strlen(phoneType))) == 0)
      {
         *outPhoneType = kMobilePhone;
      }
      else
      {
         /* default: work */
         *outPhoneType = kWorkPhone;
      }
   }
}

static const char* getPhoneType(int phoneType)
{
   switch(phoneType)
   {
      case kWorkPhone:
         return "work";
      case kFaxPhone:
         return "fax";
      case kHomePhone:
         return "home";
      case kPagerPhone:
         return "pager";
      case kMobilePhone:
         return "mobile";
   }

   return "work";
}

static void parseUserType(const char *userType, int *outUserType)
{
   if(userType && outUserType)
   {
      if(strncmp("user", userType, MIN(4, strlen(userType))) == 0)
      {
         *outUserType = kUser;
      }
      else if(strncmp("customer", userType, MIN(8, strlen(userType))) == 0)
      {
         *outUserType = kCustomer;
      }
      else
      {
         /* default: user */
         *outUserType = kUser;
      }
   }
}

static const char* getUserType(int userType)
{
   switch(userType)
   {
      case kUser:
         return "user";
      case kCustomer:
         return "customer";
   }
   return "user";
}

static void parseYesNo(const char *yesNoOpt, int *outYesNo)
{
   if(yesNoOpt && outYesNo)
   {
      if(strncmp("yes", yesNoOpt, MIN(3, strlen(yesNoOpt))) == 0)
      {
         *outYesNo = 1;
      }
      else if(strncmp("no", yesNoOpt, MIN(2, strlen(yesNoOpt))) == 0)
      {
         *outYesNo = 0;
      }
      else
      {
         /* default: no */
         *outYesNo = 0;
      }
   }
}

static const char* getYesNo(int yesNo)
{
   switch(yesNo)
   {
      case 0:
         return "no";
      case 1:
         return "yes";
   }
   return "no";
}

static void parseGlobalUser(const char *firstname, const char *initials, const char *lastname,
                            const char *username, const char *password, const char *active,
                            const char *emailAddressType, const char *emailAddress,
                            const char *phone1Type, const char *phone1, const char *phone2Type,
                            const char *phone2, const char *changePassAtLogon,
                            const char *cantChangePass, const char *userType, const char *address,
                            const char *company, const char *division, const char *department,
                            const char *security, const char *sso, const char *ttpro,
                            const char *ttpropool, const char *tcm, const char *tcmpool,
                            const char *ttrm, const char *ttrmpool, const char *scm,
                            const char *scmpool, const char *notes, GlobalUser **user)
{
   if(user)
   {
      if((*user) == NULL)
      {
         *user = malloc(sizeof(GlobalUser));
         memset(*user, 0, sizeof(GlobalUser));

         (*user)->cbSize = sizeof(GlobalUser);

         (*user)->id = (unsigned int)-1;
         (*user)->isActive = 1;
      }

      if(firstname)
      {
         free((void*)(*user)->firstName);
         (*user)->firstName = strdup(firstname);
      }
      if(initials)
      {
         free((void*)(*user)->initial);
         (*user)->initial = strdup(initials);
      }
      if(lastname)
      {
         free((void*)(*user)->lastName);
         (*user)->lastName = strdup(lastname);
      }
      if(username)
      {
         free((void*)(*user)->username);
         (*user)->username = strdup(username);
      }
      if(password)
      {
         free((void*)(*user)->password);
         (*user)->password = strdup(password);
      }
      if(active)
         parseYesNo(active, &(*user)->isActive);

      if(emailAddressType)
         parseEmailType(emailAddressType, &(*user)->emailAddressType);
      if(emailAddress)
      {
         free((void*)(*user)->emailAddress);
         (*user)->emailAddress = strdup(emailAddress);
      }
      if(phone1Type)
         parsePhoneType(phone1Type, &(*user)->phoneNumber1Type);
      if(phone2)
      {
         free((void*)(*user)->phoneNumber1);
         (*user)->phoneNumber1 = strdup(phone1);
      }
      if(phone2Type)
         parsePhoneType(phone2Type, &(*user)->phoneNumber2Type);
      if(phone2)
      {
         free((void*)(*user)->phoneNumber2);
         (*user)->phoneNumber2 = strdup(phone2);
      }
      if(changePassAtLogon)
         parseYesNo(changePassAtLogon, &(*user)->changePasswordAtLogon);
      if(cantChangePass)
         parseYesNo(cantChangePass, &(*user)->cannotChangePassword);
      if(userType)
         parseUserType(userType, &(*user)->userType);
      if(address)
      {
         free((void*)(*user)->address);
         (*user)->address = strdup(address);
      }
      if(company)
      {
         free((void*)(*user)->company);
         (*user)->company = strdup(company);
      }
      if(division)
      {
         free((void*)(*user)->division);
         (*user)->division = strdup(division);
      }
      if(department)
      {
         free((void*)(*user)->department);
         (*user)->department = strdup(department);
      }
      if(security)
         parseSecurity(security, &(*user)->securityLevel);
      if(sso)
         parseYesNo(sso, &(*user)->singleSignOn);
      if(ttpro || tcm || ttrm || scm)
      {
         parseLicensing(ttpro, ttpropool, tcm, tcmpool, ttrm, ttrmpool, scm, scmpool,
                        &(*user)->licenses, &(*user)->numLicenses);
      }
      if(notes)
      {
         free((void*)(*user)->notes);
         (*user)->notes = strdup(notes);
      }
   }
}

static GlobalUser* cloneGlobalUser(GlobalUser *inUser)
{
   int x = 0;
   GlobalUser *outUser = malloc(sizeof(GlobalUser));

   outUser->cbSize = sizeof(GlobalUser);
   outUser->id = inUser->id;

   outUser->firstName = strdup((inUser->firstName ? inUser->firstName : ""));
   outUser->initial = strdup((inUser->initial ? inUser->initial : ""));
   outUser->lastName = strdup((inUser->lastName ? inUser->lastName : ""));
   outUser->username = strdup((inUser->username ? inUser->username : ""));
   outUser->password = strdup((inUser->password ? inUser->password : ""));
   outUser->isActive = inUser->isActive;
   outUser->emailAddressType = inUser->emailAddressType;
   outUser->emailAddress = strdup((inUser->emailAddress ? inUser->emailAddress : ""));
   outUser->phoneNumber1Type = inUser->phoneNumber1Type;
   outUser->phoneNumber1 = strdup((inUser->phoneNumber1 ? inUser->phoneNumber1 : ""));
   outUser->phoneNumber2Type = inUser->phoneNumber2Type;
   outUser->phoneNumber2 = strdup((inUser->phoneNumber2 ? inUser->phoneNumber2 : ""));
   outUser->changePasswordAtLogon = inUser->changePasswordAtLogon;
   outUser->cannotChangePassword = inUser->cannotChangePassword;
   outUser->passwordNeverExpires = inUser->passwordNeverExpires:
   outUser->userType = inUser->userType;
   outUser->address = strdup((inUser->address ? inUser->address : ""));
   outUser->company = strdup((inUser->company ? inUser->company : ""));
   outUser->division = strdup((inUser->division ? inUser->division : ""));
   outUser->department = strdup((inUser->department ? inUser->department : ""));
   outUser->securityLevel = inUser->securityLevel;
   outUser->singleSignOn = inUser->singleSignOn;

   outUser->numLicenses = inUser->numLicenses;
   if(outUser->numLicenses)
   {
      outUser->licenses = calloc(outUser->numLicenses, sizeof(int));

      for(x = 0; x < outUser->numLicenses; x++)
      {
         outUser->licenses[x] = inUser->licenses[x];
      }
   }
   else
   {
      outUser->licenses = NULL;
   }

   outUser->numOldUsernames = inUser->numOldUsernames;
   if(outUser->numOldUsernames > 0)
   {
      outUser->oldUsernames = calloc(outUser->numOldUsernames, sizeof(char*));
      
      for(x = 0; x < outUser->numOldUsernames; x++)
      {
         outUser->oldUsernames[x] = strdup((inUser->oldUsernames[x] ? inUser->oldUsernames[x] : ""));
      }
   }
   else
   {
      outUser->oldUsernames = NULL;
   }

   outUser->notes = strdup((inUser->notes ? inUser->notes : ""));

   return outUser;
}

static void freeGlobalUser(GlobalUser **inUser)
{
   int x = 0;

   if(inUser && *inUser)
   {
      free((void*)(*inUser)->firstName);
      free((void*)(*inUser)->initial);
      free((void*)(*inUser)->lastName);
      free((void*)(*inUser)->username);
      free((void*)(*inUser)->password);
      free((void*)(*inUser)->emailAddress);
      free((void*)(*inUser)->phoneNumber1);
      free((void*)(*inUser)->phoneNumber2);
      free((void*)(*inUser)->address);
      free((void*)(*inUser)->company);
      free((void*)(*inUser)->division);
      free((void*)(*inUser)->department);
      free((void*)(*inUser)->licenses);

      for(x = 0; x < (*inUser)->numOldUsernames; x++)
      {
         free((*inUser)->oldUsernames[x]);
      }

      free((*inUser)->oldUsernames);
      free((void*)(*inUser)->notes);

      memset((*inUser), 0, sizeof(GlobalUser));

      free(*inUser);
      (*inUser) = NULL;
   }
}

static void parseUsers(const char *users, char ***outUsers, int *numOutUsers)
{
   char *localUsers = NULL;
   char *nextUser = NULL;
   char *context = NULL;
   int userIdx = 0;

   if(outUsers && numOutUsers)
   {
      if(users && strlen(users) > 0)
      {
         localUsers = strdup((users ? users : ""));

         *numOutUsers = 1;
         nextUser = localUsers;
         while((nextUser = strchr(nextUser, ',')) != NULL)
         {
            (*numOutUsers)++;
         }

         *outUsers = calloc((*numOutUsers), sizeof(char*));

         if((nextUser = strtok_r(localUsers, ",", &context)) != NULL)
         {
            (*outUsers)[userIdx] = strdup((nextUser ? nextUser : ""));
            userIdx++;

            while((nextUser = strtok_r(NULL, ",", &context)) != NULL)
            {
               (*outUsers)[userIdx] = strdup((nextUser ? nextUser : ""));
               userIdx++;
            }
         }
      }
      else
      {
         *numOutUsers = 0;
      }
   }

   free(localUsers);
}

static void printUser(GlobalUser *user)
{
   char *licenses = NULL;
   int x = 0;
   if(user)
   {
      printf("%u\t%s\t%s\t%s\t%s"
             "\t%s\t%s"
             "\t%s\t%s\t%s"
             "\t%s\t%s"
             "\t%s\t%s"
             "\t%s\t%s\t%s\t%s"
             "\t%s\t%s\t%s"
             "\t%s"
             "\t%s",
             user->id, user->username, user->firstName, user->initial, user->lastName,
             (user->isActive == 0 ? "inactive" : "active"), getEmailType(user->emailAddressType),
             (user->emailAddress), getPhoneType(user->phoneNumber1Type), user->phoneNumber1,
             getPhoneType(user->phoneNumber2Type), user->phoneNumber2,
             getYesNo(user->changePasswordAtLogon), getYesNo(user->cannotChangePassword),
             getUserType(user->userType), user->address, user->company, user->division,
             user->department, getSecurity(user->securityLevel), getYesNo(user->singleSignOn),
             (licenses = getLicenses(user->licenses, user->numLicenses)),
             (user->ldapServer && strlen(user->ldapServer) > 0 ? user->ldapServer : "<none>"));

      /* old user names */
      printf("\t");
      if(user->oldUsernames && user->numOldUsernames > 0)
      {
         for(x = 0; x < user->numOldUsernames - 1; x++)
         {
            printf("%s,", user->oldUsernames[x]);
         }

         printf(user->oldUsernames[user->numOldUsernames - 1]);
      }

      /* notes */
      printf("\t\"%s\"\n", user->notes);
   }

   if(licenses)
   {
      free(licenses);
   }
}

static void printUsers(GlobalUser *users[], int numUsers)
{
   int x = 0;
   /* print header */
   printf("id\tusername\tfirst name\tinitials\tlast name\tactivation\temail type\temail address\t"
          "phone number 1 type\tphone number 1\tphone number 2 type\t phone number 2\t"
          "must change password at next login\tcannot change password\tuser type\taddress\t"
          "company\tdivision\tdepartment\tsecurity level\tsingle sign on allowed\tlicenses\t"
          "ldapserver\toldusernames\tnotes\n");

   for(x = 0; x < numUsers; x++)
   {
      printUser(users[x]);
   }
}

static const char* getProduct(int product)
{
   static const char *szTTPro = "ttpro";
   static const char *szTCM = "tcm";
   static const char *szTTRM = "ttrm";
   static const char *szSCM = "scm";
   static const char *szBlank = "";

   switch(product)
   {
      case kTTP:
         return szTTPro;
      case kSCM:
         return szSCM;
      case kTCM:
         return szTCM;
      case kRM:
         return szTTRM;
   }

   return szBlank;
}

static const char* getType(int type)
{
   static const char *szFloating = "floating";
   static const char *szNamed = "named";
   static const char *szBlank = "";

   switch(type)
   {
      case kFloating:
         return szFloating;
      case kNamed:
         return szNamed;
   }

   return szBlank;
}

static const char* getVersion(int version)
{
   static const char *szMaint = "Maintenance";
   static const char *szUnprintable = "Unprintable";
   static char szVersion[4];

   if(version == 0)
   {
      return szMaint;
   }
   else
   {
      memset(szVersion, 0, 4 * sizeof(char));
      /* currently only 3 digit versions are supported */
      assert(version < 1000);
      if(version > 999)
      {
         return szUnprintable;
      }
      else
      {
         sprintf(szVersion, "%d", version);
         return (const char*)szVersion;
      }
   }
}

static void printSerialNumber(SerialNumber *serialNumber)
{
   int x = 0;
   if(serialNumber)
   {
      printf("%u\t%s\t%s\t"
             "%s\t%s\t"
             "%s\t%s\t"
             "%s\t%s\t"
             "%s\t",
             serialNumber->id, serialNumber->serialNumber, getProduct(serialNumber->product),
             getType(serialNumber->type), getYesNo(serialNumber->isEnabled),
             getYesNo(serialNumber->isExpired), getYesNo(serialNumber->isEval),
             getVersion(serialNumber->version), serialNumber->serialNumberDate,
             serialNumber->expireDate);

      if(serialNumber->maintenanceExtenders && serialNumber->numMaintenanceExtenders > 0)
      {
         for(x = 0; x < serialNumber->numMaintenanceExtenders - 1; x++)
         {
            printf("%s,", serialNumber->maintenanceExtenders[x]);
         }

         printf("%s",
                serialNumber->maintenanceExtenders[serialNumber->numMaintenanceExtenders - 1]);
      }
      printf("\n");
   }
}

static void printSerialNumbers(SerialNumber *serialNumbers[], int numSerialNumbers)
{
   int x = 0;

   /* print header */
   printf("id\tserialNumber\tproduct\ttype\tis enabled\tis expired\tis evaluation\tversion\t"
          "serial number date\texpiration date\tmaintenance extenders\n");

   for(x = 0; x < numSerialNumbers; x++)
   {
      printSerialNumber(serialNumbers[x]);
   }
}

static void printPool(Pool *pool)
{
   int x = 0;
   if(pool)
   {
      printf("%u\t%s\t%s\t",
             pool->id, pool->name, getProduct(pool->product));

      if(pool->serialNumbers && pool->numSerialNumbers > 0)
      {
         for(x = 0; x < pool->numSerialNumbers - 1; x++)
         {
            printf("%u:%d,", pool->serialNumbers[x]->snID, pool->serialNumbers[x]->count);
         }

         printf("%u:%d", pool->serialNumbers[pool->numSerialNumbers - 1]->snID,
                pool->serialNumbers[pool->numSerialNumbers - 1]->count);
      }
      printf("\n");
   }
}

static void printPools(Pool *pools[], int numPools)
{
   int x = 0;

   /* print header */
   printf("id\tname\tproduct\tserial number counts\n");

   for(x = 0; x < numPools; x++)
   {
      printPool(pools[x]);
   }
}

static int parseProduct(char *product)
{
   int prod = 0;
   int productLen = strlen(product);

   if(strncmp("ttpro", product, MIN(5, productLen)) == 0)
   {
      prod = kTTP;
   }
   else if(strncmp("tcm", product, MIN(3, productLen)) == 0)
   {
      prod = kTCM;
   }
   else if(strncmp("ttrm", product, MIN(4, productLen)) == 0)
   {
      prod = kRM;
   }
   else if(strncmp("scm", product, MIN(3, productLen)) == 0)
   {
      prod = kSCM;
   }

   return prod;
}

static Pool* clonePool(Pool *pool, char *name, char *product)
{
   Pool *newPool = NULL;
   int x = 0;

   if(pool)
   {
      newPool = calloc(1, sizeof(Pool));
      newPool->cbSize = sizeof(Pool);
      newPool->name = strdup((name && strlen(name) > 0 ? name : pool->name));
      newPool->product = (product && strlen(product) > 0 ? parseProduct(product) : pool->product);

      if(pool->serialNumbers && pool->numSerialNumbers > 0)
      {
         newPool->serialNumbers = calloc(pool->numSerialNumbers, sizeof(SNCount*));
         for(x = 0; x < pool->numSerialNumbers; x++)
         {
            newPool->serialNumbers[x] = calloc(1, sizeof(SNCount));
            newPool->serialNumbers[x]->cbSize = sizeof(SNCount);
            newPool->serialNumbers[x]->snID = pool->serialNumbers[x]->snID;
            newPool->serialNumbers[x]->count = pool->serialNumbers[x]->count;
         }
      }
      else
      {
         newPool->serialNumbers = NULL;
         newPool->numSerialNumbers = 0;
      }
   }

   return newPool;
}

static SNCount** cloneCountsSkip(SNCount **counts, int numCounts, unsigned int skipID)
{
   SNCount **newCounts = calloc(numCounts - 1, sizeof(SNCount*));
   int x = 0;
   int y = 0;

   for(x = 0; x < numCounts && y < numCounts - 1; x++)
   {
      if(counts[x] && counts[x]->snID != skipID)
      {
         newCounts[y] = calloc(1, sizeof(SNCount));
         newCounts[y]->cbSize = sizeof(SNCount);
         newCounts[y]->snID = counts[x]->snID;
         newCounts[y]->count = counts[x]->count;

         y++;
      }
   }

   return newCounts;
}

static SNCount** cloneCountsAdd(SNCount **counts, int numCounts, unsigned int addID, int addCount)
{
   SNCount **newCounts = calloc(numCounts + 1, sizeof(SNCount*));
   int x = 0;

   for(x = 0; x < numCounts; x++)
   {
      newCounts[x] = calloc(1, sizeof(SNCount));
      newCounts[x]->cbSize = sizeof(SNCount);
      newCounts[x]->snID = counts[x]->snID;
      newCounts[x]->count = counts[x]->count;
   }

   newCounts[numCounts] = calloc(1, sizeof(SNCount));
   newCounts[numCounts]->cbSize = sizeof(SNCount);
   newCounts[numCounts]->snID = addID;
   newCounts[numCounts]->count = addCount;

   return newCounts;
}

static void freeCounts(SNCount **counts, int numCounts)
{
   int x = 0;

   if(counts && numCounts > 0)
   {
      for(x = 0; x < numCounts; x++)
      {
         free(counts[x]);
         counts[x] = NULL;
      }
   }
   
   free(counts);
}

static void freePool(Pool *pool)
{
   freeCounts(pool->serialNumbers, pool->numSerialNumbers);
   free(pool->name);
   free(pool);
}

static int importUser(const char *lsLoginInfo, const char *userName, const char *name,
                      const char *ldapServer, const char *securityLevel, const char *ttpro,
                      const char *tcm, const char *ttrm, const char *scm)
{
   LSAdminHandle *admin = NULL;
   int returnCode = 0;
   char *lsusername = NULL;
   char *lspassword = NULL;
   char *lshostname = NULL;
   int portNum = 0;
   int security = 0;
   int numLicenses = 0;
   int *licenses = NULL;
   LicenseAssignment **licenseAssignments = NULL;
   int x = 0;

   returnCode = parseLoginInfo(lsLoginInfo, &lsusername, &lspassword, &lshostname, &portNum);

   if(returnCode == 0) returnCode = CreateLSAdmin(&admin);
   if(returnCode == 0) returnCode = LSAdminConnect(admin, lshostname, portNum, lsusername,
                                                   lspassword);
   if(returnCode == 0)
   {
      parseSecurity(securityLevel, &security);
      parseLicensing(ttpro, NULL, tcm, NULL, ttrm, NULL, scm, NULL, &licenseAssignments,
                     &numLicenses);

      if(numLicenses > 0)
      {
         licenses = (int*)calloc(numLicenses, sizeof(int));

         for(x = 0; x < numLicenses; x++)
         {
            licenses[x] = licenseAssignments[x]->licenseType;

            free(licenseAssignments[x]);
            licenseAssignments[x] = NULL;
         }

         free(licenseAssignments);
      }

      if(licenses)
      {
         returnCode = LSAdminImportLDAPUser(admin, name, userName, ldapServer, security, licenses,
                                            numLicenses);
      }
      else
      {
         returnCode = APIERROR_INVALIDARGS;
      }

      LSAdminDisconnect(admin);
   }

   DestroyLSAdmin(admin);
   
   free(lsusername);
   free(lspassword);
   free(lshostname);

   return returnCode;
}

static int addUserEx(const char *lsLoginInfo, GlobalUser *newUser)
{
   LSAdminHandle *admin = NULL;
   int returnCode = 0;
   char *lsusername = NULL;
   char *lspassword = NULL;
   char *lshostname = NULL;
   int portNum = 0;

   returnCode = parseLoginInfo(lsLoginInfo, &lsusername, &lspassword, &lshostname, &portNum);

   if(returnCode == 0) returnCode = CreateLSAdmin(&admin);
   if(returnCode == 0) returnCode = LSAdminConnect(admin, lshostname, portNum, lsusername,
                                                   lspassword);
   if(returnCode == 0)
   {
      if(newUser)
      {
         returnCode = LSAdminAddUserEx(admin, newUser);
      }
      else
      {
         returnCode = APIERROR_INVALIDARGS;
      }

      LSAdminDisconnect(admin);
   }

   DestroyLSAdmin(admin);
   
   free(lsusername);
   free(lspassword);
   free(lshostname);

   return returnCode;
}

static int editUser(const char *lsLoginInfo, const char *username, const char *firstname,
                    const char *initials, const char *lastname, const char *newusername,
                    const char *password, const char *active, const char *emailAddressType,
                    const char *emailAddress, const char *phone1Type, const char *phone1,
                    const char *phone2Type, const char *phone2, const char *changePassAtLogon,
                    const char *cantChangePass, const char *userType, const char *address,
                    const char *company, const char *division, const char *department,
                    const char *security, const char *sso, const char *ttpro, const char *ttpropool,
                    const char *tcm, const char *tcmpool, const char *ttrm, const char *ttrmpool,
                    const char *scm, const char *scmpool, const char *notes)
{
   LSAdminHandle *admin = NULL;
   int returnCode = 0;
   char *lsusername = NULL;
   char *lspassword = NULL;
   char *lshostname = NULL;
   int portNum = 0;
   const char *usersToList[1];
   GlobalUser **globalUsers = NULL;
   int numGlobalUsers = 0;
   GlobalUser *user = NULL;

   usersToList[0] = username;

   returnCode = parseLoginInfo(lsLoginInfo, &lsusername, &lspassword, &lshostname, &portNum);

   if(returnCode == 0) returnCode = CreateLSAdmin(&admin);
   if(returnCode == 0) returnCode = LSAdminConnect(admin, lshostname, portNum, lsusername,
                                                   lspassword);
   if(returnCode == 0)
   {
      returnCode = LSAdminGetUsers(admin, usersToList, 1, &globalUsers, &numGlobalUsers);

      if(returnCode == 0)
      {
         /* clone retrieved user */
         if(numGlobalUsers == 1)
         {
            user = cloneGlobalUser(globalUsers[0]);

            if(user)
            {
               parseGlobalUser(firstname, initials, lastname, newusername, password, active,
                               emailAddressType, emailAddress, phone1Type, phone1, phone2Type,
                               phone2, changePassAtLogon, cantChangePass, userType, address,
                               company, division, department, security, sso, ttpro, ttpropool, tcm,
                               tcmpool, ttrm, ttrmpool, scm, scmpool, notes, &user);

               returnCode = LSAdminEditUser(admin, username, user);

               freeGlobalUser(&user);
            }
            else
            {
               returnCode = APIERROR_INVALIDARGS;
            }
         }
         else
         {
            returnCode = APIERROR_GENERALAPIERROR;
         }
      }

      LSAdminFreeGlobalUsers(globalUsers, numGlobalUsers);
      LSAdminDisconnect(admin);
   }

   DestroyLSAdmin(admin);
   
   free(lsusername);
   free(lspassword);
   free(lshostname);

   return returnCode;
}

static int associateUser(const char *lsLoginInfo, const char *username, const char *ldapUsername,
                         const char *ldapServer)
{
   LSAdminHandle *admin = NULL;
   int returnCode = 0;
   char *lsusername = NULL;
   char *lspassword = NULL;
   char *lshostname = NULL;
   int portNum = 0;

   returnCode = parseLoginInfo(lsLoginInfo, &lsusername, &lspassword, &lshostname, &portNum);

   if(returnCode == 0) returnCode = CreateLSAdmin(&admin);
   if(returnCode == 0) returnCode = LSAdminConnect(admin, lshostname, portNum, lsusername,
                                                   lspassword);
   if(returnCode == 0)
   {
      if(username && ldapServer)
      {
         returnCode = LSAdminAssociateToLDAPUser(admin, username,
                                                 (ldapUsername ? ldapUsername : username),
                                                 ldapServer);
      }
      else
      {
         returnCode = APIERROR_INVALIDARGS;
      }

      LSAdminDisconnect(admin);
   }

   DestroyLSAdmin(admin);
   
   free(lsusername);
   free(lspassword);
   free(lshostname);

   return returnCode;
}

static int listUsers(const char *lsLoginInfo, char *users)
{
   LSAdminHandle *admin = NULL;
   int returnCode = 0;
   char *lsusername = NULL;
   char *lspassword = NULL;
   char *lshostname = NULL;
   int portNum = 0;
   char **usersToList = NULL;
   int numUsersToList = 0;
   GlobalUser **globalUsers = NULL;
   int numGlobalUsers = 0;

   returnCode = parseLoginInfo(lsLoginInfo, &lsusername, &lspassword, &lshostname, &portNum);

   if(returnCode == 0) returnCode = CreateLSAdmin(&admin);
   if(returnCode == 0) returnCode = LSAdminConnect(admin, lshostname, portNum, lsusername,
                                                   lspassword);
   if(returnCode == 0)
   {
      if(users)
      {
         parseUsers(users, &usersToList, &numUsersToList);
      }

      returnCode = LSAdminGetUsers(admin, (const char **)usersToList, numUsersToList, &globalUsers,
                                   &numGlobalUsers);

      if(numGlobalUsers > 0)
      {
         printUsers(globalUsers, numGlobalUsers);
      }

      LSAdminFreeGlobalUsers(globalUsers, numGlobalUsers);
      LSAdminDisconnect(admin);
   }

   DestroyLSAdmin(admin);
   
   free(lsusername);
   free(lspassword);
   free(lshostname);

   return returnCode;
}

static int addSN(const char *lsLoginInfo, char *serialNumber)
{
   LSAdminHandle *admin = NULL;
   int returnCode = 0;
   char *lsusername = NULL;
   char *lspassword = NULL;
   char *lshostname = NULL;
   int portNum = 0;

   returnCode = parseLoginInfo(lsLoginInfo, &lsusername, &lspassword, &lshostname, &portNum);

   if(returnCode == 0) returnCode = CreateLSAdmin(&admin);
   if(returnCode == 0) returnCode = LSAdminConnect(admin, lshostname, portNum, lsusername,
                                                   lspassword);
   if(returnCode == 0)
   {
      returnCode = LSAdminAddSerialNumber(admin, (const char*)serialNumber);
      LSAdminDisconnect(admin);
   }

   DestroyLSAdmin(admin);
   
   free(lsusername);
   free(lspassword);
   free(lshostname);

   return returnCode;
}

static int listSNs(const char *lsLoginInfo)
{
   LSAdminHandle *admin = NULL;
   int returnCode = 0;
   char *lsusername = NULL;
   char *lspassword = NULL;
   char *lshostname = NULL;
   int portNum = 0;
   SerialNumber **serialNumbers = NULL;
   int numSerialNumbers = 0;
   returnCode = parseLoginInfo(lsLoginInfo, &lsusername, &lspassword, &lshostname, &portNum);

   if(returnCode == 0) returnCode = CreateLSAdmin(&admin);
   if(returnCode == 0) returnCode = LSAdminConnect(admin, lshostname, portNum, lsusername,
                                                   lspassword);
   if(returnCode == 0)
   {
      returnCode = LSAdminGetSerialNumbers(admin, &serialNumbers, &numSerialNumbers);

      if(numSerialNumbers > 0)
      {
         printSerialNumbers(serialNumbers, numSerialNumbers);
      }

      LSAdminFreeSerialNumbers(serialNumbers, numSerialNumbers);
      LSAdminDisconnect(admin);
   }

   DestroyLSAdmin(admin);
   
   free(lsusername);
   free(lspassword);
   free(lshostname);

   return returnCode;
}

static int deleteSN(const char *lsLoginInfo, char *id)
{
   LSAdminHandle *admin = NULL;
   int returnCode = 0;
   char *lsusername = NULL;
   char *lspassword = NULL;
   char *lshostname = NULL;
   int portNum = 0;
   unsigned int snId = 0;

   returnCode = parseLoginInfo(lsLoginInfo, &lsusername, &lspassword, &lshostname, &portNum);

   if(returnCode == 0) returnCode = CreateLSAdmin(&admin);
   if(returnCode == 0) returnCode = LSAdminConnect(admin, lshostname, portNum, lsusername,
                                                   lspassword);
   if(returnCode == 0)
   {
      sscanf(id, "%u", &snId);
      returnCode = LSAdminDeleteSerialNumber(admin, snId);
      LSAdminDisconnect(admin);
   }

   DestroyLSAdmin(admin);
   
   free(lsusername);
   free(lspassword);
   free(lshostname);

   return returnCode;
}

static int addPool(const char *lsLoginInfo, char *name, char *product)
{
   LSAdminHandle *admin = NULL;
   int returnCode = 0;
   char *lsusername = NULL;
   char *lspassword = NULL;
   char *lshostname = NULL;
   int portNum = 0;
   Pool *pool = NULL;

   returnCode = parseLoginInfo(lsLoginInfo, &lsusername, &lspassword, &lshostname, &portNum);

   if(returnCode == 0) returnCode = CreateLSAdmin(&admin);
   if(returnCode == 0) returnCode = LSAdminConnect(admin, lshostname, portNum, lsusername,
                                                   lspassword);
   if(returnCode == 0)
   {
      pool = (Pool*)calloc(1, sizeof(Pool));
      pool->id = (unsigned int)-1;
      pool->name = strdup(name);
      pool->product = parseProduct(product);
      pool->serialNumbers = NULL;
      pool->numSerialNumbers = 0;

      returnCode = LSAdminAddPool(admin, pool);

      free(pool->name);
      free(pool);

      LSAdminDisconnect(admin);
   }

   DestroyLSAdmin(admin);
   
   free(lsusername);
   free(lspassword);
   free(lshostname);

   return returnCode;
}

static int listPools(const char *lsLoginInfo)
{
   LSAdminHandle *admin = NULL;
   int returnCode = 0;
   char *lsusername = NULL;
   char *lspassword = NULL;
   char *lshostname = NULL;
   int portNum = 0;
   Pool **pools = NULL;
   int numPools = 0;

   returnCode = parseLoginInfo(lsLoginInfo, &lsusername, &lspassword, &lshostname, &portNum);

   if(returnCode == 0) returnCode = CreateLSAdmin(&admin);
   if(returnCode == 0) returnCode = LSAdminConnect(admin, lshostname, portNum, lsusername,
                                                   lspassword);
   if(returnCode == 0)
   {
      returnCode = LSAdminGetPools(admin, &pools, &numPools);

      if(pools && numPools > 0)
      {
         printPools(pools, numPools);
      }

      LSAdminDisconnect(admin);
   }

   DestroyLSAdmin(admin);
   
   free(lsusername);
   free(lspassword);
   free(lshostname);

   return returnCode;
}

static int editPool(const char *lsLoginInfo, char *id, char *name, char *product)
{
   LSAdminHandle *admin = NULL;
   int returnCode = 0;
   char *lsusername = NULL;
   char *lspassword = NULL;
   char *lshostname = NULL;
   int portNum = 0;
   Pool **pools = NULL;
   int numPools = 0;
   int x = 0;
   Pool *editPool = NULL;
   unsigned int poolID = 0;

   returnCode = parseLoginInfo(lsLoginInfo, &lsusername, &lspassword, &lshostname, &portNum);

   if(returnCode == 0) returnCode = CreateLSAdmin(&admin);
   if(returnCode == 0) returnCode = LSAdminConnect(admin, lshostname, portNum, lsusername,
                                                   lspassword);
   if(returnCode == 0)
   {
      sscanf(id, "%u", &poolID);
      returnCode = LSAdminGetPools(admin, &pools, &numPools);

      if(returnCode == 0)
      {
         /* find the specified pool */
         for(x = 0; x < numPools && editPool == NULL; x++)
         {
            if(pools[x] && pools[x]->id  == poolID)
            {
               editPool = clonePool(pools[x], name, product);;
            }
         }

         if(!editPool)
         {
            fprintf(stderr, "Specified pool does not exist in License Server\n");
            returnCode = APIERROR_INVALIDARGS;
         }
         else
         {
            returnCode = LSAdminEditPools(admin, &editPool, 1);
            freePool(editPool);
            editPool = NULL;
         }
      }

      LSAdminFreePools(pools, numPools);
      LSAdminDisconnect(admin);
   }

   DestroyLSAdmin(admin);
   
   free(lsusername);
   free(lspassword);
   free(lshostname);

   return returnCode;
}

static int moveLicenses(const char *lsLoginInfo, char *fromid, char *toid, char *snid, char *count)
{
   LSAdminHandle *admin = NULL;
   int returnCode = 0;
   char *lsusername = NULL;
   char *lspassword = NULL;
   char *lshostname = NULL;
   int portNum = 0;
   Pool **pools = NULL;
   int numPools = 0;
   int x = 0;
   unsigned int fromPoolID = 0;
   unsigned int toPoolID = 0;
   const int FROM = 0;
   const int TO = 1;
   Pool *newPools[2];
   unsigned int serialNumberID = 0;
   int numLicenses = 0;
   int actionFlag = 0; /* 0 == inline adjustment, 1 == rebuild list, 2 == invalid count, 3 == not found */
   SNCount **newCounts = NULL;
   SNCount **tmpCounts = NULL;

   newPools[FROM] = NULL;
   newPools[TO] = NULL;

   returnCode = parseLoginInfo(lsLoginInfo, &lsusername, &lspassword, &lshostname, &portNum);

   if(returnCode == 0) returnCode = CreateLSAdmin(&admin);
   if(returnCode == 0) returnCode = LSAdminConnect(admin, lshostname, portNum, lsusername,
                                                   lspassword);
   if(returnCode == 0)
   {
      sscanf(fromid, "%u", &fromPoolID);
      sscanf(toid, "%u", &toPoolID);
      returnCode = LSAdminGetPools(admin, &pools, &numPools);

      if(returnCode == 0)
      {
         /* find the specified pools */
         for(x = 0; x < numPools && newPools[FROM] == NULL && newPools[TO] == NULL; x++)
         {
            if(pools[x])
            {
               if(pools[x]->id == fromPoolID)
               {
                  newPools[FROM] = clonePool(pools[x], NULL, NULL);
               }
               else if(pools[x]->id == toPoolID)
               {
                  newPools[TO] = clonePool(pools[x], NULL, NULL);
               }
            }
         }

         if(!newPools[FROM] || !newPools[TO])
         {
            fprintf(stderr, "Specified pool does not exist in License Server\n");
            returnCode = APIERROR_INVALIDARGS;
         }
         else
         {
            sscanf(snid, "%u", &serialNumberID);
            sscanf(count, "%d", &numLicenses);

            /* assume the item is not in the list, and that's an error */
            actionFlag = 3;
            for(x = 0; x < newPools[FROM]->numSerialNumbers && actionFlag == 3; x++)
            {
               if(newPools[FROM]->serialNumbers[x] && newPools[FROM]->serialNumbers[x]->snID == serialNumberID)
               {
                  if(newPools[FROM]->serialNumbers[x]->count > numLicenses)
                  {
                     newPools[FROM]->serialNumbers[x]->count -= numLicenses;
                     actionFlag = 0;
                  }
                  else if(newPools[FROM]->serialNumbers[x]->count == numLicenses)
                  {
                     /* rebuild list without this element */
                     actionFlag = 1;
                  }
                  else
                  {
                     /* too many licenses requested... error */
                     actionFlag = 2;
                  }
               }
            }

            if(actionFlag == 1)
            {
               newCounts = cloneCountsSkip(newPools[FROM]->serialNumbers, newPools[FROM]->numSerialNumbers,
                                           serialNumberID);
               tmpCounts = newPools[FROM]->serialNumbers;
               newPools[FROM]->serialNumbers = newCounts;
               freeCounts(tmpCounts, newPools[FROM]->numSerialNumbers);
               newPools[FROM]->numSerialNumbers--;
            }
            else if(actionFlag == 2)
            {
               fprintf(stderr, "Invalid serial number count specified\n");
               returnCode = APIERROR_INVALIDARGS;
            }
            else if(actionFlag == 3)
            {
               fprintf(stderr, "Specified serial number is not in source pool\n");
               returnCode = APIERROR_INVALIDARGS;
            }

            if(returnCode == 0)
            {
               /* assume the item is not in the list and it needs rebuilt */
               actionFlag = 1;
               for(x = 0; x < newPools[TO]->numSerialNumbers; x++)
               {
                  if(newPools[TO]->serialNumbers[x] && newPools[TO]->serialNumbers[x]->snID == serialNumberID)
                  {
                     newPools[TO]->serialNumbers[x]->count += numLicenses;
                  }
               }

               if(actionFlag == 1)
               {
                  newCounts = cloneCountsAdd(newPools[TO]->serialNumbers, newPools[TO]->numSerialNumbers,
                                             serialNumberID, numLicenses);
                  tmpCounts = newPools[TO]->serialNumbers;
                  newPools[TO]->serialNumbers = newCounts;
                  freeCounts(tmpCounts, newPools[TO]->numSerialNumbers);
                  newPools[TO]->numSerialNumbers++;
               }

               returnCode = LSAdminEditPools(admin, pools, 2);
            }
         }

         freePool(newPools[FROM]);
         freePool(newPools[TO]);
         newPools[FROM] = NULL;
         newPools[TO] = NULL;
      }

      LSAdminFreePools(pools, numPools);
      LSAdminDisconnect(admin);
   }

   DestroyLSAdmin(admin);
   
   free(lsusername);
   free(lspassword);
   free(lshostname);

   return returnCode;
}

static int deletePool(const char *lsLoginInfo, char *id)
{
   LSAdminHandle *admin = NULL;
   int returnCode = 0;
   char *lsusername = NULL;
   char *lspassword = NULL;
   char *lshostname = NULL;
   int portNum = 0;
   unsigned int poolId = 0;

   returnCode = parseLoginInfo(lsLoginInfo, &lsusername, &lspassword, &lshostname, &portNum);

   if(returnCode == 0) returnCode = CreateLSAdmin(&admin);
   if(returnCode == 0) returnCode = LSAdminConnect(admin, lshostname, portNum, lsusername,
                                                   lspassword);
   if(returnCode == 0)
   {
      sscanf(id, "%u", &poolId);
      returnCode = LSAdminDeletePool(admin, poolId);
      LSAdminDisconnect(admin);
   }

   DestroyLSAdmin(admin);
   
   free(lsusername);
   free(lspassword);
   free(lshostname);

   return returnCode;
}

int main(int argc, char *argv[])
{
   /* control variables */
   int returnCode = 0;
   int nextOpt = -1;
   int abort = 0;

   /* char*s to grab option values */
   char *userName = NULL;
   char *name = NULL;
   char *ldapServer = NULL;
   char *ttpro = NULL;
   char *tcm = NULL;
   char *ttrm = NULL;
   char *scm = NULL;
   char *firstName = NULL;
   char *initials = NULL;
   char *lastName = NULL;
   char *password = NULL;
   char *email = NULL;
   char *phone1 = NULL;
   char *address = NULL;
   char *company = NULL;
   char *division = NULL;
   char *department = NULL;
   char *securityLevel = NULL;
   char *active = NULL;
   char *emailtype = NULL;
   char *phone1type = NULL;
   char *phone2type = NULL;
   char *phone2 = NULL;
   char *changepassword = NULL;
   char *cantchangepassword = NULL;
   char *type = NULL;
   char *sso = NULL;
   char *notes = NULL;
   char *ldapusername = NULL;
   char *newusername = NULL;
   char *users = NULL;
   char *ttpropool = NULL;
   char *tcmpool = NULL;
   char *ttrmpool = NULL;
   char *scmpool = NULL;
   char *serialnumber = NULL;
   char *id = NULL;
   char *product = NULL;
   char *fromid = NULL;
   char *toid = NULL;
   char *snid = NULL;
   char *count = NULL;

   /* char*s to keep command info */
   char *lsLoginInfo = NULL;
   char *lsapiCommand = NULL;

   /* command-line option definition */
   const char* const shortOptions = "hV";
   const struct option longOptions[] = {
      {"username", required_argument, NULL, 2},
      {"name", required_argument, NULL, 3},
      {"ldapserver", required_argument, NULL, 4},
      {"ttpro", required_argument, NULL, 5},
      {"tcm", required_argument, NULL, 6},
      {"ttrm", required_argument, NULL, 7},
      {"scm", required_argument, NULL, 8},
      {"firstname", required_argument, NULL, 9},
      {"initials", required_argument, NULL, 10},
      {"lastname", required_argument, NULL, 11},
      {"password", required_argument, NULL, 12},
      {"email", required_argument, NULL, 13},
      {"phone1", required_argument, NULL, 14},
      {"address", required_argument, NULL, 15},
      {"company", required_argument, NULL, 16},
      {"division", required_argument, NULL, 17},
      {"department", required_argument, NULL, 18},
      {"security", required_argument, NULL, 19},
      {"active", required_argument, NULL, 20},
      {"emailtype", required_argument, NULL, 21},
      {"phone1type", required_argument, NULL, 22},
      {"phone2type", required_argument, NULL, 23},
      {"phone2", required_argument, NULL, 24},
      {"forcechangepassword", required_argument, NULL, 25},
      {"preventchangepassword", required_argument, NULL, 26},
      {"type", required_argument, NULL, 27},
      {"singlesignon", required_argument, NULL, 28},
      {"notes", required_argument, NULL, 29},
      {"lsusername", required_argument, NULL, 30},
      {"ldapusername", required_argument, NULL, 31},
      {"newusername", required_argument, NULL, 32},
      {"users", required_argument, NULL, 33},
      {"ttpropool", required_argument, NULL, 34},
      {"tcmpool", required_argument, NULL, 35},
      {"ttrmpool", required_argument, NULL, 36},
      {"scmpool", required_argument, NULL, 37},
      {"serialnumber", required_argument, NULL, 38},
      {"id", required_argument, NULL, 39},
      {"product", required_argument, NULL, 40},
      {"fromid", required_argument, NULL, 41},
      {"toid", required_argument, NULL, 42},
      {"snid", required_argument, NULL, 43},
      {"count", required_argument, NULL, 44},
      {"help", no_argument, NULL, 'h'},
      {"version", no_argument, NULL, 'V'},
      {0, 0, 0, 0}
   };

   /* user object to populate */
   GlobalUser *user = NULL;

   int commandLen = 0;

   /* hang on to the program name */
   programName = argv[0];

   /* parse all command-line options */
   do
   {
      nextOpt = getopt_long(argc, argv, shortOptions, longOptions, NULL);

      switch(nextOpt)
      {
         case 2:
            userName = optarg;
            break;
         case 3:
            name = optarg;
            break;
         case 4:
            ldapServer = optarg;
            break;
         case 5:
            ttpro = optarg;
            break;
         case 6:
            tcm = optarg;
            break;
         case 7:
            ttrm = optarg;
            break;
         case 8:
            scm = optarg;
            break;
         case 9:
            firstName = optarg;
            break;
         case 10:
            initials = optarg;
            break;
         case 11:
            lastName = optarg;
            break;
         case 12:
            password = optarg;
            break;
         case 13:
            email = optarg;
            break;
         case 14:
            phone1 = optarg;
            break;
         case 15:
            address = optarg;
            break;
         case 16:
            company = optarg;
            break;
         case 17:
            division = optarg;
            break;
         case 18:
            department = optarg;
            break;
         case 19:
            securityLevel = optarg;
            break;
         case 20:
            active = optarg;
            break;
         case 21:
            emailtype = optarg;
            break;
         case 22:
            phone1type = optarg;
            break;
         case 23:
            phone2type = optarg;
            break;
         case 24:
            phone2 = optarg;
            break;
         case 25:
            changepassword = optarg;
            break;
         case 26:
            cantchangepassword = optarg;
            break;
         case 27:
            type = optarg;
            break;
         case 28:
            sso = optarg;
            break;
         case 29:
            notes = optarg;
            break;
         case 30:
            userName = optarg;
            break;
         case 31:
            ldapusername = optarg;
            break;
         case 32:
            newusername = optarg;
            break;
         case 33:
            users = optarg;
            break;
         case 34:
            ttpropool = optarg;
            break;
         case 35:
            tcmpool = optarg;
            break;
         case 36:
            ttrmpool = optarg;
            break;
         case 37:
            scmpool = optarg;
            break;
         case 38:
            serialnumber = optarg;
            break;
         case 39:
            id = optarg;
            break;
         case 40:
            product = optarg;
            break;
         case 41:
            fromid = optarg;
            break;
         case 42:
            toid = optarg;
            break;
         case 43:
            snid = optarg;
            break;
         case 44:
            count = optarg;
            break;
         case 'h':
            /* -h or --help */
            printUsage(stdout);
            abort = 1;
            break;
         case 'V':
            /* -V or --version */
            printf("%s: Helix ALM License Server Command Line Tool 1.1.0\n", programName);
            abort = 1;
            break;
         case '?':
            /* unrecognized command */
            printUsage(stderr);
            returnCode = 1;
            abort = 1;
            break;
      }
   } while(nextOpt != -1 && abort == 0);

   /* options parsed... unless we've aborted, move on to the login information and command... */
   while(abort == 0 && optind < argc)
   {
      if(lsLoginInfo == NULL)
      {
         lsLoginInfo = argv[optind++];
      }
      else if(lsapiCommand == NULL)
      {
         lsapiCommand = argv[optind++];
      }
      else
      {
         /* ruh-roh reorge... we've got too many arguments */
         returnCode = 1;
         abort = 1;
         fprintf(stderr, "%s: Unrecognized parameter: %s\n", programName, argv[optind]);
      }
   }

   /* options parsed, and we don't need to abort... */
   if(abort == 0)
   {
      if(!lsLoginInfo)
      {
         fprintf(stderr, "%s: No License Server specified\n", programName);
      }
      else if(!lsapiCommand)
      {
         fprintf(stderr, "%s: No command given\n", programName);
      }
      else
      {
         commandLen = strlen(lsapiCommand);
         /* execute the command! */
         if(strncmp("importuser", lsapiCommand, MIN(10, commandLen)) == 0)
         {
            returnCode = importUser(lsLoginInfo, userName, name, ldapServer, securityLevel, ttpro,
                                    tcm, ttrm, scm);
         }
         else if(strncmp("associateuser", lsapiCommand, MIN(13, commandLen)) == 0)
         {
            returnCode = associateUser(lsLoginInfo, userName, ldapusername, ldapServer);
         }
         else if(strncmp("adduser", lsapiCommand, MIN(7, commandLen)) == 0)
         {
            parseGlobalUser(firstName, initials, lastName, userName, password, active, emailtype,
                            email, phone1type, phone1, phone2type, phone2, changepassword,
                            cantchangepassword, type, address, company, division, department,
                            securityLevel, sso, ttpro, ttpropool, tcm, tcmpool, ttrm, ttrmpool, scm,
                            scmpool, notes, &user);
            returnCode = addUserEx(lsLoginInfo, user);
            freeGlobalUser(&user);
            free(user);
         }
         else if(strncmp("listusers", lsapiCommand, MIN(9, commandLen)) == 0)
         {
            returnCode = listUsers(lsLoginInfo, users);
         }
         else if(strncmp("edituser", lsapiCommand, MIN(8, commandLen)) == 0)
         {
            returnCode = APIERROR_GENERALAPIERROR;

            returnCode = editUser(lsLoginInfo, userName, firstName, initials, lastName, newusername,
                                  password, active, emailtype, email, phone1type, phone1,
                                  phone2type, phone2, changepassword, cantchangepassword, type,
                                  address, company, division, department, securityLevel, sso, ttpro,
                                  ttpropool, tcm, tcmpool, ttrm, ttrmpool, scm, scmpool, notes);
         }
         else if(strncmp("addsn", lsapiCommand, MIN(5, commandLen)) == 0)
         {
            returnCode = APIERROR_GENERALAPIERROR;
            returnCode = addSN(lsLoginInfo, serialnumber);
         }
         else if(strncmp("listsns", lsapiCommand, MIN(7, commandLen)) == 0)
         {
            returnCode = APIERROR_GENERALAPIERROR;
            returnCode = listSNs(lsLoginInfo);
         }
         else if(strncmp("deletesn", lsapiCommand, MIN(8, commandLen)) == 0)
         {
            returnCode = APIERROR_GENERALAPIERROR;
            returnCode = deleteSN(lsLoginInfo, id);
         }
         else if(strncmp("addpool", lsapiCommand, MIN(7, commandLen)) == 0)
         {
            returnCode = APIERROR_GENERALAPIERROR;
            returnCode = addPool(lsLoginInfo, name, product);
         }
         else if(strncmp("listpools", lsapiCommand, MIN(9, commandLen)) == 0)
         {
            returnCode = APIERROR_GENERALAPIERROR;
            returnCode = listPools(lsLoginInfo);
         }
         else if(strncmp("editpool", lsapiCommand, MIN(8, commandLen)) == 0)
         {
            returnCode = APIERROR_GENERALAPIERROR;
            returnCode = editPool(lsLoginInfo, id, name, product);
         }
         else if(strncmp("movelicenses", lsapiCommand, MIN(12, commandLen)) == 0)
         {
            returnCode = APIERROR_GENERALAPIERROR;
            returnCode = moveLicenses(lsLoginInfo, fromid, toid, snid, count);
         }
         else if(strncmp("deletepool", lsapiCommand, MIN(10, commandLen)) == 0)
         {
            returnCode = APIERROR_GENERALAPIERROR;
            returnCode = deletePool(lsLoginInfo, id);
         }
         else
         {
            fprintf(stderr, "%s: Unrecognized command: %s\n", programName, lsapiCommand);
         }
      }
   }

   return returnCode;
}

/* eof */


